{
  "type": "record",
  "name": "DataHubAccessTokenKey",
  "namespace": "com.linkedin.pegasus2avro.metadata.key",
  "doc": "Key for a DataHub Access Token",
  "fields": [
    {
      "name": "id",
      "type": "string",
      "doc": "Access token's SHA-256 hashed JWT signature"
    }
  ],
  "Aspect": {
    "name": "dataHubAccessTokenKey"
  }
}