{
  "type": "record",
  "name": "MLHyperParam",
  "namespace": "com.linkedin.pegasus2avro.ml.metadata",
  "doc": "Properties associated with an ML Hyper Param",
  "fields": [
    {
      "name": "name",
      "type": "string",
      "doc": "Name of the MLHyperParam"
    },
    {
      "name": "description",
      "type": [
        "null",
        "string"
      ],
      "doc": "Documentation of the MLHyperParam",
      "default": null
    },
    {
      "name": "value",
      "type": [
        "null",
        "string"
      ],
      "doc": "The value of the MLHyperParam",
      "default": null
    },
    {
      "name": "createdAt",
      "type": [
        "null",
        "long"
      ],
      "doc": "Date when the MLHyperParam was developed",
      "default": null
    }
  ],
  "Aspect": {
    "name": "mlHyperParam"
  }
}